import os
import unittest
from itertools import product
from pathlib import Path

from tablemap import Conn


class TestCore(unittest.TestCase):
    def setUp(self):
        self.dbfile = 'test/sample.db'
        self.dbfile1 = 'test/sample1.db'

        self.t1 = [
            {'col1': 'a', 'col2': 4},
            {'col1': 'a', 'col2': 5},
            {'col1': 'b', 'col2': 1},
            {'col1': 'c', 'col2': 3},
            {'col1': 'c', 'col2': 4},
            {'col1': 'c', 'col2': 7},
            {'col1': 'd', 'col2': 2},

        ]

        self.t2 = [
            {'col1': 'd', 'col3': 3},
            {'col1': 'b', 'col3': 1},
            {'col1': 'e', 'col3': 3},
            {'col1': 'e', 'col3': 4},
            {'col1': 'd', 'col3': 2},
        ]

    def tearDown(self):
        if Path(self.dbfile).is_file():
            os.remove(self.dbfile)
        if Path(self.dbfile1).is_file():
            os.remove(self.dbfile1)

    def test_do_later(self):
        conn = Conn(self.dbfile)
        conn['t1'] = self.t1

        xs, ys = conn['t1'].tee(2)
        xs.by('col1')
        tot = sum(ys['col2'])
        xs.fold(
            col3=lambda rs: sum(rs['col2']),
            col4=tot
        )
        self.assertEqual(xs.list(),
                         [{'col1': 'a', 'col3': 9, 'col4': 26},
                          {'col1': 'b', 'col3': 1, 'col4': 26},
                          {'col1': 'c', 'col3': 14, 'col4': 26},
                          {'col1': 'd', 'col3': 2, 'col4': 26}])

    def test_zip(self):
        conn = Conn(self.dbfile)
        conn['t1'] = self.t1

        def tempfn(rs):
            return rs.zip([
                {'n': 1},
                {'n': 2}
            ])

        conn['t1_1'] = conn['t1'].by('col1').map(tempfn)
        self.assertEqual(conn['t1_1'].list(),
                         [{'col1': 'a', 'col2': 4, 'n': 1},
                          {'col1': 'a', 'col2': 5, 'n': 2},
                          {'col1': 'b', 'col2': 1, 'n': 1},
                          {'col1': 'c', 'col2': 3, 'n': 1},
                          {'col1': 'c', 'col2': 4, 'n': 2},
                          {'col1': 'd', 'col2': 2, 'n': 1}]
                         )

    def test_tee_n_new(self):
        conn = Conn(self.dbfile)

        def tempfn(rs):
            rs1, rs2 = rs.tee(2)
            return rs1.order('col2 desc').chain(rs2.order('col2'))

        conn['t1'] = self.t1
        conn['t1_1'] = conn['t1'].by('col1').map(tempfn)
        self.assertEqual(conn['t1_1']['col2'],
                         [5, 4, 4, 5, 1, 1, 7, 4, 3, 3, 4, 7, 2, 2])

        def tempfn(rs):
            rs1, rs2 = rs.tee(2)
            rs1.update(col2=lambda r: r['col2'] + 1)
            return rs1.order('col2 desc').chain(rs2.order('col2'))

        conn['t1'] = self.t1
        conn['t1_1'] = conn['t1'].by('col1').map(tempfn)
        self.assertEqual(conn['t1_1']['col2'],
                         [6, 5, 5, 6, 2, 2, 8, 5, 4, 4, 5, 8, 3, 3])

        def tempfn(rs):
            rs1, rs2 = rs.tee(2)
            rs1.update(col2=lambda r: r['col2'] + 1)
            return rs1.order('col2 desc').chain(rs2.new().order('col2'))

        conn['t1'] = self.t1
        conn['t1_1'] = conn['t1'].by('col1').map(tempfn)
        self.assertEqual(conn['t1_1']['col2'],
                         [6, 5, 4, 5, 2, 1, 8, 5, 4, 3, 4, 7, 3, 2])

    def test_tee_n_new2(self):
        conn = Conn(self.dbfile)

        def tempfn(rs):
            rs1, rs2 = rs.tee(2)
            return rs1.order('col2').zip(rs2.order('col2 desc')
                                         .rename(col2_1='col2'))

        conn['t1'] = self.t1
        conn['t1_1'] = conn['t1'].by('col1').map(tempfn)
        xs = [{'col1': 'a', 'col2': 4, 'col2_1': 5},
              {'col1': 'a', 'col2': 5, 'col2_1': 4},
              {'col1': 'b', 'col2': 1, 'col2_1': 1},
              {'col1': 'c', 'col2': 3, 'col2_1': 7},
              {'col1': 'c', 'col2': 4, 'col2_1': 4},
              {'col1': 'c', 'col2': 7, 'col2_1': 3},
              {'col1': 'd', 'col2': 2, 'col2_1': 2}]
        self.assertEqual(conn['t1_1'].list(), xs)

        def tempfn1(rs):
            rs1, rs2 = rs.tee(2)
            rs1.update(col2=lambda r: r['col2'] + 1)
            return rs1.order('col2').zip(rs2.order('col2 desc')
                                         .rename(col2_1='col2'))

        conn['t1_1'] = conn['t1'].by('col1').map(tempfn1)

        xs = [{'col1': 'a', 'col2': 5, 'col2_1': 6},
              {'col1': 'a', 'col2': 6, 'col2_1': 5},
              {'col1': 'b', 'col2': 2, 'col2_1': 2},
              {'col1': 'c', 'col2': 4, 'col2_1': 8},
              {'col1': 'c', 'col2': 5, 'col2_1': 5},
              {'col1': 'c', 'col2': 8, 'col2_1': 4},
              {'col1': 'd', 'col2': 3, 'col2_1': 3}]
        self.assertEqual(conn['t1_1'].list(), xs)

        def tempfn2(rs):
            rs1, rs2 = rs.tee(2)
            rs1.update(col2=lambda r: r['col2'] + 1)
            return rs1.order('col2').zip(rs2.new().order('col2 desc')
                                         .rename(col2_1='col2'))

        conn['t1_1'] = conn['t1'].by('col1').map(tempfn2)

        xs = [{'col1': 'a', 'col2': 5, 'col2_1': 5},
              {'col1': 'a', 'col2': 6, 'col2_1': 4},
              {'col1': 'b', 'col2': 2, 'col2_1': 1},
              {'col1': 'c', 'col2': 4, 'col2_1': 7},
              {'col1': 'c', 'col2': 5, 'col2_1': 4},
              {'col1': 'c', 'col2': 8, 'col2_1': 3},
              {'col1': 'd', 'col2': 3, 'col2_1': 2}]
        self.assertEqual(conn['t1_1'].list(), xs)

    def test_index(self):
        conn = Conn(self.dbfile)
        conn['t1'] = self.t1

        def cnt():
            n = 0

            def innerfn(rs):
                nonlocal n
                n += 1
                return rs.update(n=n)
            return innerfn

        conn['t1_1'] = conn['t1'].by('col1').map(cnt())
        self.assertEqual(
            conn['t1_1'].list(), [{'col1': 'a', 'col2': 4, 'n': 1},
                                  {'col1': 'a', 'col2': 5, 'n': 1},
                                  {'col1': 'b', 'col2': 1, 'n': 2},
                                  {'col1': 'c', 'col2': 3, 'n': 3},
                                  {'col1': 'c', 'col2': 4, 'n': 3},
                                  {'col1': 'c', 'col2': 7, 'n': 3},
                                  {'col1': 'd', 'col2': 2, 'n': 4}])

    def test_fold1(self):
        conn = Conn(self.dbfile)

        def minmax(col):
            def fn(rs):
                xs = [r[col] for r in rs]
                return [min(xs), max(xs)]
            return fn

        conn['t1'] = self.t1
        conn['t1_1'] = conn['t1'].by('col1').fold(
            sum=lambda rs: sum(rs['col2']),
            mm0=lambda rs: min(rs['col2']),
            mm1=lambda rs: max(rs['col2'])
        )

        self.assertEqual(conn['t1_1'].list(),
                         [{'col1': 'a', 'sum': 9, 'mm0': 4, 'mm1': 5},
                          {'col1': 'b', 'sum': 1, 'mm0': 1, 'mm1': 1},
                          {'col1': 'c', 'sum': 14, 'mm0': 3, 'mm1': 7},
                          {'col1': 'd', 'sum': 2, 'mm0': 2, 'mm1': 2}])

        conn['t1_1'] = conn['t1'].windowed(4, 3).fold(
            sum=lambda rs: sum(rs['col2']),
            mm0=lambda rs: min(rs['col2']),
            mm1=lambda rs: max(rs['col2'])
        )
        self.assertEqual(conn['t1_1'].list(),
                         [{'sum': 13, 'mm0': 1, 'mm1': 5},
                          {'sum': 16, 'mm0': 2, 'mm1': 7}
                          ])

        conn['t1_1'] = conn['t1'].windowed(4, 5).fold(
            sum=lambda rs: sum(rs['col2']),
            mm0=lambda rs: min(rs['col2']),
            mm1=lambda rs: max(rs['col2'])
        ).rename(
            mm1='mm0',
            mm0='mm1'
        )
        self.assertEqual(conn['t1_1'].list(),
                         [{'sum': 13, 'mm1': 1, 'mm0': 5},
                          {'sum': 9, 'mm1': 2, 'mm0': 7}
                          ])

    def test_by(self):
        def sum_by_col1(rs):
            tot = sum(rs['col2'])
            return rs.update(col2=tot).islice(1)

        conn = Conn(self.dbfile)
        conn['t1'] = self.t1
        conn['t1_sum'] = conn['t1'].by('col1')\
            .map(sum_by_col1)

        self.assertEqual(list(conn['t1_sum'].iter()),
                         [{'col1': 'a', 'col2': 9},
                          {'col1': 'b', 'col2': 1},
                          {'col1': 'c', 'col2': 14},
                          {'col1': 'd', 'col2': 2}]
                         )
        conn['t1_sum2'] = conn['t1']\
            .by('col1').map(sum_by_col1)

        self.assertEqual(conn['t1_sum'].list(), conn['t1_sum2'].list())

    def test_chain(self):
        conn = Conn(self.dbfile)
        conn['t1'] = self.t1
        conn['t1_double'] = conn['t1'].chain(conn['t1'])
        self.assertEqual(conn['t1'].list() + conn['t1'].list(),
                         conn['t1_double'].list())

        conn['t1_double'] = conn['t1'].chain(self.t1)
        self.assertEqual(conn['t1'].list() + conn['t1'].list(),
                         conn['t1_double'].list())

    def test_df(self):
        import pandas as pd
        conn = Conn(self.dbfile)
        conn['t1'] = self.t1
        df = pd.DataFrame(conn['t1'].iter())
        conn['t1_copy'] = df.to_dict('records')
        self.assertEqual(conn['t1'].list(), conn['t1_copy'].list())

    def test_by_after_map(self):
        def sumit(rs):
            tot = sum(rs['col2'])
            return rs.update(tot=tot)

        def count(rs):
            n = len(rs.list())
            return rs.update(cnt=n)

        conn = Conn(self.dbfile)
        conn['t1'] = self.t1
        conn['t1_sum'] = conn['t1'].by('col1')\
            .map(sumit).by('tot').map(count)

        # on another db
        conn1 = Conn(self.dbfile1)
        conn1['t1_sum'] = conn['t1'].by('col1').map(sumit).by('tot').map(count)
        conn1['t1_sum2'] = conn['t1'].by('col1').map(sumit)\
            .by('tot').map(count).iter()

        self.assertEqual(conn['t1_sum'].list(), conn1['t1_sum'].list())
        self.assertEqual(conn['t1_sum'].list(), conn1['t1_sum2'].list())

    def test_multiple_by(self):

        conn = Conn(self.dbfile)
        conn['t1'] = self.t1
        conn['t1_1'] = conn['t1'].update(col3=1).by('col1', 'col3')\
            .map(lambda rs: rs)

        conn['t1_2'] = conn['t1'].update(col3=1).by('col1').map(lambda rs: rs)
        self.assertEqual(conn['t1_1'].list(), conn['t1_2'].list())

        conn['t1_1_double'] = conn['t1_1']\
            .chain(conn['t1'].update(col3=1).by('col1', 'col3')
                   .map(lambda rs: rs))
        conn['t1_1_double2'] = conn['t1_1'].chain(conn['t1_1'])

        self.assertEqual(conn['t1_1_double'].list(),
                         conn['t1_1_double2'].list())

    def test_islice(self):

        conn = Conn(self.dbfile)
        conn['t1'] = self.t1
        conn['t2'] = self.t2

        self.assertEqual(len(conn['t1'].islice(3).list()), 3)
        self.assertEqual(len(conn['t1'].islice(3, None).list()), 4)
        self.assertEqual(len(conn['t1'].islice(0, None, 3).list()), 3)

        xs = conn['t1'].iter()
        ys = conn['t2'].iter()

        conn['t1_2'] = ({**x, **y} for x, y in product(xs, ys))

        n1 = len(conn['t1'].list())
        n2 = len(conn['t2'].list())
        n3 = len(conn['t1_2'].list())

        self.assertEqual(n3, n1 * n2)

    def test_takewhile(self):
        conn = Conn(self.dbfile)
        conn['t1'] = self.t1

        self.assertEqual(len(conn['t1']
                             .takewhile(lambda r: r['col2'] > 1).list()), 2)

        self.assertEqual(len(conn['t1']
                             .order('col2 desc')
                             .takewhile(lambda r: r['col2'] > 1).list()), 6)

    def test_dropwhile(self):
        conn = Conn(self.dbfile)
        conn['t1'] = self.t1

        self.assertEqual(len(conn['t1']
                             .dropwhile(lambda r: r['col2'] > 1).list()), 5)

        self.assertEqual(len(conn['t1']
                             .order('col2')
                             .dropwhile(lambda r: r['col2'] < 4).list()), 4)

    def test_windowed(self):
        conn = Conn(self.dbfile)
        conn['t1'] = self.t1

        def add1(rs):
            return rs.update(col2=lambda r: r['col2'] + 1)

        self.assertEqual(conn['t1'].windowed(4, 2).map(add1).list(),
                         [{'col1': 'a', 'col2': 5},
                          {'col1': 'a', 'col2': 6},
                          {'col1': 'b', 'col2': 2},
                          {'col1': 'c', 'col2': 4},
                          {'col1': 'b', 'col2': 2},
                          {'col1': 'c', 'col2': 4},
                          {'col1': 'c', 'col2': 5},
                          {'col1': 'c', 'col2': 8},
                          {'col1': 'c', 'col2': 5},
                          {'col1': 'c', 'col2': 8},
                          {'col1': 'd', 'col2': 3}])

    def test_filter(self):
        conn = Conn(self.dbfile)

        conn['t1'] = self.t1
        conn['t1_1'] = conn['t1'].filter(lambda r: r['col2'] > 2)

        self.assertEqual(len(conn['t1_1'].list()), 5)

    def test_update(self):
        conn = Conn(self.dbfile)
        conn['t1'] = self.t1
        conn['t1_1'] = conn['t1'].update(
            col2=lambda r: r['col2'] + 1,
            col3=lambda r: r['col1'] + str(r['col2'])
        )
        # table is fetched as it's inserted. not sure if it's a feature.
        self.assertEqual(conn['t1_1'].list()[0],
                         {'col1': 'a', 'col2': 5, 'col3': 'a5'})

    def test_select(self):
        conn = Conn(self.dbfile)
        conn['t1'] = self.t1

        # for r in conn['t1'].iter():
        #     print(r)

        conn['t1_1'] = conn['t1'].update(
            col2=lambda r: r['col2'] + 1,
            col3=lambda r: r['col1'] + str(r['col2'])
        )

        r = next(conn['t1_1'].select('col1', 'col3').iter())
        self.assertEqual(len(r), 2)

        # test deselect
        r = next(conn['t1_1'].deselect('col1', 'col3').iter())
        self.assertEqual(len(r), 1)

    def test_join(self):
        conn = Conn(self.dbfile)
        conn['t1'] = self.t1
        conn['t2'] = self.t2

        conn['t1_1'] = conn['t1'].by('col1').inner_join(conn['t2'].by('col1'))

        self.assertEqual(len(conn['t1_1'].list()), 3)

        conn['t1_2'] = conn['t1'].by('col1').left_join(conn['t2'].by('col1'))

        self.assertEqual(len(conn['t1_2'].list()), 8)

        conn['t1_3'] = conn['t1'].by('col1').right_join(conn['t2'].by('col1'))

        self.assertEqual(len(conn['t1_3'].list()), 5)

        conn['t1_4'] = conn['t1'].by('col1').full_join(conn['t2'].by('col1'))

        self.assertEqual(conn['t1_4'].list(), [
            {'col1': 'a', 'col2': 4, 'col3': ''},
            {'col1': 'a', 'col2': 5, 'col3': ''},
            {'col1': 'b', 'col2': 1, 'col3': 1},
            {'col1': 'c', 'col2': 3, 'col3': ''},
            {'col1': 'c', 'col2': 4, 'col3': ''},
            {'col1': 'c', 'col2': 7, 'col3': ''},
            {'col1': 'd', 'col2': 2, 'col3': 3},
            {'col1': 'd', 'col2': 2, 'col3': 2},
            {'col1': 'e', 'col2': '', 'col3': 3},
            {'col1': 'e', 'col2': '', 'col3': 4}]
        )

    def test_join2(self):
        conn = Conn(self.dbfile)
        conn['t1'] = self.t1
        conn['t2'] = self.t2

        conn['t1_1'] = conn['t1'].chain([{'col1': 'd', 'col2': 9}, {
            'col1': 'd', 'col2': 19}]).by('col1')\
            .full_join(conn['t2'].by('col1'))

        self.assertEqual(conn['t1_1'].list(),
                         [{'col1': 'a', 'col2': 4, 'col3': ''},
                          {'col1': 'a', 'col2': 5, 'col3': ''},
                          {'col1': 'b', 'col2': 1, 'col3': 1},
                          {'col1': 'c', 'col2': 3, 'col3': ''},
                          {'col1': 'c', 'col2': 4, 'col3': ''},
                          {'col1': 'c', 'col2': 7, 'col3': ''},
                          {'col1': 'd', 'col2': 2, 'col3': 3},
                          {'col1': 'd', 'col2': 2, 'col3': 2},
                          {'col1': 'd', 'col2': 9, 'col3': 3},
                          {'col1': 'd', 'col2': 9, 'col3': 2},
                          {'col1': 'd', 'col2': 19, 'col3': 3},
                          {'col1': 'd', 'col2': 19, 'col3': 2},
                          {'col1': 'e', 'col2': '', 'col3': 3},
                          {'col1': 'e', 'col2': '', 'col3': 4}])

    def test_distinct(self):
        conn = Conn(self.dbfile)
        conn['t1'] = self.t1
        conn['t1_1'] = conn['t1'].distinct('col1')
        self.assertEqual(len(conn['t1_1'].list()), 4)
        conn['t1_2'] = conn['t1'].distinct('col1', 'col2')
        self.assertEqual(len(conn['t1_2'].list()), 7)

    def test_order(self):
        conn = Conn(self.dbfile)
        conn['t1'] = self.t1

        [g1, g2, g3, g4] = [list(xs) for _, xs in conn['t1'].order(
            'col1', 'col2 desc').group('col1').iter()]
        self.assertEqual(
            g1, [{'col1': 'a', 'col2': 5}, {'col1': 'a', 'col2': 4}]
        )
        self.assertEqual(g2, [{'col1': 'b', 'col2': 1}])
        self.assertEqual(g3, [{'col1': 'c', 'col2': 7}, {
                         'col1': 'c', 'col2': 4}, {'col1': 'c', 'col2': 3}])
        self.assertEqual(g4, [{'col1': 'd', 'col2': 2}])

    def test_readme(self):
        t1 = [
            {'col1': 'a', 'col2': 4},
            {'col1': 'a', 'col2': 5},
            {'col1': 'b', 'col2': 1},
        ]

        conn = Conn(self.dbfile)
        conn['t1'] = t1

        for r1, r2 in zip(conn['t1'].list(), conn['t1'].iter()):
            self.assertEqual(r1, r2)

        import pandas as pd

        df = pd.DataFrame(conn['t1'].iter())
        conn['t1_copy'] = df.to_dict('records')

        self.assertEqual(conn['t1'].list(), conn['t1_copy'].list())

        conn['t1_1'] = conn['t1']\
            .filter(lambda r: r['col2'] > 2)\
            .update(col2=lambda r: r['col2'] + 1)

        self.assertEqual(conn['t1_1'].list()[0]['col2'], 5)

        conn['t1_col2sum_groupby_col1'] = conn['t1'].by('col1')\
            .fold(col2_sum=lambda rs: sum(rs['col2']))\
            .deselect('col2')

        conn['t1_col2sum_groupby_col1_1'] = conn['t1']\
            .order('col1', 'col2 desc').group('col1')\
            .fold(col2_sum=lambda rs: sum(rs['col2']))\
            .deselect('col2')

        self.assertEqual(conn['t1_col2sum_groupby_col1'].list(),
                         conn['t1_col2sum_groupby_col1_1'].list())

        conn['t1_double'] = conn['t1'].chain(conn['t1'])
        self.assertEqual(len(conn['t1_double'].list()),
                         2 * len(conn['t1'].list()))

        conn['t1_double'] = conn['t1'].chain(t1)

        self.assertEqual(len(conn['t1_double'].list()),
                         2 * len(conn['t1'].list()))

        conn['t1_1'] = conn['t1']\
            .filter(lambda r: r['col2'] > 2)\
            .update(
                col2=lambda r: r['col2'] + 1,
                col3=lambda r: r['col1'] + str(r['col2'])
        )
        self.assertEqual(conn['t1_1'].list(),
                         [{'col1': 'a', 'col2': 5, 'col3': 'a5'},
                          {'col1': 'a', 'col2': 6, 'col3': 'a6'}])

        conn['t1_1'] = conn['t1'].update(col3=lambda r: r['col2'] + 1)\
            .deselect('col1', 'col2')

        self.assertEqual(conn['t1_1'].list(),
                         [{'col3': 5}, {'col3': 6}, {'col3': 2}])

        conn['t1_1'] = conn['t1'].distinct('col1')
        self.assertEqual(conn['t1_1'].list(), [
                         {'col1': 'a', 'col2': 4}, {'col1': 'b', 'col2': 1}])

        conn['t2'] = [
            {'col1': 'b', 'col3': -1},
            {'col1': 'c', 'col3': 3},
            {'col1': 'b', 'col3': ''},
        ]

        conn['t1_col3'] = conn['t1'].by('col1')\
            .full_join(conn['t2'].by('col1'))

        self.assertEqual(conn['t1_col3'].list(),
                         [{'col1': 'a', 'col2': 4, 'col3': ''},
                         {'col1': 'a', 'col2': 5, 'col3': ''},
                         {'col1': 'b', 'col2': 1, 'col3': -1},
                         {'col1': 'b', 'col2': 1, 'col3': ''},
                          {'col1': 'c', 'col2': '', 'col3': 3}]
                         )



"""A data wrangling tool which requires no knowledge of Pandas or SQL.
"""

import os
import signal
import sqlite3
import tempfile
from collections import deque
from contextlib import contextmanager
from itertools import (chain, dropwhile, groupby, islice, product, takewhile,
                       tee)
from pathlib import Path, PurePath


class Conn:
    """Connection to a SQL database file.

    Example:
        conn = Conn('sample.db')
    """

    def __init__(self, dbfile):
        # dbfile must be a filename(str), can't be :memory:
        if PurePath(dbfile).is_absolute():
            self._dbfile = dbfile
        else:
            self._dbfile = os.path.join(os.getcwd(), dbfile)

        self._dbconn = None

    def __getitem__(self, tname):
        return Rows((self, tname))

    def __setitem__(self, tname, val):
        with _connect(self._dbfile) as c:
            self._dbconn = c
            _delete(c, tname)
            _insert(c, tname, _rows2iter(val))


class Rows:
    """
    """

    def __init__(self, src):
        if isinstance(src, tuple):
            conn, tname = src

            self._conn = conn
            self._elts = None
            self._genfns = [_build_initgen(conn, tname, [])]

            self._insts = [{
                'cmd': 'fetch',
                'tname': tname,
            }]

        else:
            self._conn = None
            self._elts = list(src)

            def gen():
                yield from self._elts

            self._genfns = [gen]
            self._insts = [{
                'cmd': 'elts'
            }]

    def __str__(self):
        return "\n".join(str(r) for r in self.iter())

    def __getitem__(self, col):
        return [r[col] for r in self.iter()]

    def zip(self, other):
        pgen = self._genfns[-1]

        def gen():
            for r1, r2 in zip(pgen(), _rows2iter(other)):
                r1.update(r2)
                yield r1

        self._genfns.append(gen)
        self._insts.append({
            'cmd': 'zip',
            'other': other
        })
        return self

    def chain(self, other):
        pgen = self._genfns[-1]

        def gen():
            for r in pgen():
                yield r
            for r in _rows2iter(other):
                yield r

        self._genfns.append(gen)
        self._insts.append({
            'cmd': 'chain',
            'other': other
        })
        return self

    # each column may contain asc or desc
    # ex) 'col1 desc', 'col2 asc'
    def order(self, *cols):
        if len(self._insts) == 1 and self._insts[-1]['cmd'] == 'fetch':
            tname = self._insts[-1]['tname']
            # replace the previous generator
            self._genfns[0] = _build_initgen(self._conn, tname, cols)
            self._insts.append({
                'cmd': 'order',
                'cols': cols
            })
            return self

        pgen = self._genfns[-1]

        def gen_from_sql():
            try:
                tmpdbfd, tmpdb = tempfile.mkstemp()
                with _connect(tmpdb) as c:
                    _insert(c, 'temp', pgen())
                    yield from _fetch(c, 'temp', cols)
            finally:
                # must close the file descriptor to delete it
                os.close(tmpdbfd)
                if Path(tmpdb).is_file():
                    os.remove(tmpdb)

        def gen_simp():
            xs = list(pgen())
            for col in reversed(cols):
                rev = False
                cs = col.split()
                if len(cs) == 2:
                    c, desc = cs
                    if desc.lower() == 'desc':
                        rev = True
                else:
                    c = cs[0]
                xs.sort(key=lambda r: r[c], reverse=rev)
            yield from xs

        if self._insts[0]['cmd'] == 'fetch':
            self._genfns.append(gen_from_sql)
        else:
            self._genfns.append(gen_simp)

        self._genfns.append(gen_from_sql)
        self._insts.append({
            'cmd': 'order',
            'cols': cols
        })
        return self

    def apply(self, fn):
        pgen = self._genfns[-1]

        def gen_single():
            for r in pgen():
                yield from fn(r)

        def gen_group():
            for k, rs in pgen():
                yield from fn(k, rs)

        def gen_windowed():
            for rs in pgen():
                yield from fn(rs)

        if self._insts[-1]['cmd'] == 'group':
            self._genfns.append(gen_group)
        elif self._insts[-1]['cmd'] == 'windowed':
            self._genfns.append(gen_windowed)
        else:
            self._genfns.append(gen_single)

        self._insts.append({
            'cmd': 'apply',
            'fn': fn
        })
        return self

    def group(self, *cols):
        pgen = self._genfns[-1]

        def gen():
            yield from groupby(pgen(), _keyfn(cols))

        self._genfns.append(gen)
        self._insts.append({
            'cmd': 'group',
            'cols': cols
        })
        return self

    def _merge(self, fn, other):
        pgen = self._genfns[-1]

        def gen():
            yield from _step(fn, pgen(), _rows2iter(other))

        self._genfns.append(gen)
        self._insts.append({
            'cmd': '_merge',
            'fn': fn,
            'other': other
        })
        return self

    def map(self, fn):
        if self._insts[-1]['cmd'] == 'group':
            def _fn2gen(f):
                def gen(_k, rs):
                    x = f(Rows(rs))
                    if isinstance(x, dict):
                        yield x
                    elif isinstance(x, Rows):
                        yield from x.iter()
                return gen
            return self.apply(_fn2gen(fn))

        elif self._insts[-1]['cmd'] == 'windowed':
            def _fn2gen1(f):

                def gen(rs):
                    x = f(Rows(rs))
                    if isinstance(x, dict):
                        yield x
                    elif isinstance(x, Rows):
                        yield from x.iter()
                return gen

            def _fn2gen2(f):
                def gen(rs):
                    rs = rs[:step] + [r.copy() for r in rs[step:]]
                    x = f(Rows(rs))
                    if isinstance(x, dict):
                        yield x
                    elif isinstance(x, Rows):
                        yield from x.iter()
                return gen

            step = self._insts[-1]['step']
            n = self._insts[-1]['n']

            if step >= n:
                # no need to copy
                return self.apply(_fn2gen1(fn))
            return self.apply(_fn2gen2(fn))

        # single row
        def _fn2gen_single(f):
            def gen(r):
                x = f(r)
                if isinstance(x, dict):
                    yield x
                elif isinstance(x, Rows):
                    yield from x.iter()
            return gen
        return self.apply(_fn2gen_single(fn))

    def by(self, *cols):
        # need to cut out 'desc', 'asc'
        cols_for_group = [col.split()[0] for col in cols]
        return self.order(*cols).group(*cols_for_group)

    def filter(self, pred):
        def gen(r):
            if pred(r):
                yield r

        return self.apply(gen)

    def update(self, **kwargs):
        def updatefn(r):
            for k, v in kwargs.items():
                if callable(v):
                    r.update({k: v(r)})
                else:
                    r.update({k: v})
            yield r

        return self.apply(updatefn)

    # pretty expensive for what it actually does
    # but this version does not depend on the order of rows.
    def rename(self, **kwargs):
        kwargs_rev = {oldkey: newkey for newkey, oldkey in kwargs.items()}

        def renamefn(r):
            yield {kwargs_rev.get(oldkey, oldkey): v
                   for oldkey, v in r.items()}

        return self.apply(renamefn)

    def fold(self, **kwargs):
        gcols = []
        if self._insts[-1]['cmd'] == 'group':
            gcols = self._insts[-1]['cols']

        def foldfn1(rs):
            ds = Rows(rs)
            r = {}
            for k, fn in kwargs.items():
                if callable(fn):
                    r[k] = fn(ds)
                else:
                    r[k] = fn
            yield r

        def foldfn2(keys, rs):
            ds = Rows(rs)
            r = dict(zip(gcols, keys))
            for k, fn in kwargs.items():
                if callable(fn):
                    r[k] = fn(ds)
                else:
                    r[k] = fn
            yield r

        if self._insts[-1]['cmd'] == 'windowed':
            return self.apply(foldfn1)
        if self._insts[-1]['cmd'] == 'group':
            return self.apply(foldfn2)
        raise ValueError("fold must be preceded by grouping processes")

    def select(self, *cols):
        def selectfn(r):
            yield {col: r[col] for col in cols}

        return self.apply(selectfn)

    def deselect(self, *cols):
        cols = set(cols)

        def deselectfn(r):
            yield {k: v for k, v in r.items() if k not in cols}

        return self.apply(deselectfn)

    def inner_join(self, other):
        def innerjoin_fn(rs1, rs2, _left, _right):
            if rs1 != [] and rs2 != []:
                for r1, r2 in product(rs1, rs2):
                    r1.update(r2)
                    yield r1

        return self._merge(innerjoin_fn, other)

    def left_join(self, other):
        def leftjoin_fn(rs1, rs2, _left, right):
            if rs1 != [] and rs2 != []:
                for r1, r2 in product(rs1, rs2):
                    r1.update(r2)
                    yield r1

            elif rs1 != [] and rs2 == []:
                for r1 in rs1:
                    for col in right:
                        r1[col] = ''
                    yield r1

        return self._merge(leftjoin_fn, other)

    def right_join(self, other):
        def rightjoin_fn(rs1, rs2, left, _right):
            if rs1 != [] and rs2 != []:
                for r1, r2 in product(rs1, rs2):
                    r2.update(r1)
                    yield r2

            elif rs1 == [] and rs2 != []:
                for r2 in rs2:
                    for col in left:
                        r2[col] = ''
                    yield r2

        return self._merge(rightjoin_fn, other)

    def full_join(self, other):
        def fulljoin_fn(rs1, rs2, left, right):
            if rs1 != [] and rs2 != []:
                for r1, r2 in product(rs1, rs2):
                    # TODO: updating on the same object multiple times.
                    # Still, why works?!!,
                    r1.update(r2)
                    yield r1

            elif rs1 != [] and rs2 == []:
                for r1 in rs1:
                    for col in right:
                        r1[col] = ''
                    yield r1

            elif rs1 == [] and rs2 != []:
                for r2 in rs2:
                    d = {col: '' for col in left}
                    d.update(r2)
                    yield d

        return self._merge(fulljoin_fn, other)

    def distinct(self, *columns):
        def distinctfn(_k, rs):
            yield next(rs)

        return self.by(*columns).apply(distinctfn)

    def windowed(self, n, step):
        pgen = self._genfns[-1]

        def gen():
            yield from _windowed(pgen(), n, step)

        self._genfns.append(gen)
        self._insts.append({
            'cmd': 'windowed',
            'n': n,
            'step': step
        })
        return self

    def islice(self, *args):
        pgen = self._genfns[-1]

        def gen():
            yield from islice(pgen(), *args)

        self._genfns.append(gen)
        self._insts.append({
            'cmd': 'islice',
            'args': args
        })
        return self

    def takewhile(self, pred):
        pgen = self._genfns[-1]

        def gen():
            yield from takewhile(pred, pgen())

        self._genfns.append(gen)
        self._insts.append({
            'cmd': 'takewhile',
            'pred': pred
        })
        return self

    def dropwhile(self, pred):
        pgen = self._genfns[-1]

        def gen():
            yield from dropwhile(pred, pgen())

        self._genfns.append(gen)
        self._insts.append({
            'cmd': 'dropwhile',
            'pred': pred
        })
        return self

    def iter(self):
        # TODO: analyze self._insts and check if it's semantically ok.
        yield from self._genfns[-1]()

    def list(self):
        return list(self.iter())

    # experimental
    def tee(self, n):
        return tuple(Rows(x) for x in tee(self.iter(), n))

    # experimental
    def new(self):
        return Rows([r.copy() for r in self.iter()])


def _insert_statement(name, d):
    """insert into foo values (:a, :b, :c, ...)

    Notice the colons.
    """
    keycols = ', '.join(":" + c.strip() for c in d)
    return f"insert into {name} values ({keycols})"


def _create_statement(tname, cols):
    """Create table if not exists foo (...)

    Note:
        Every type is numeric.
    """
    schema = ', '.join([col + ' ' + 'numeric' for col in cols])
    return f"create table if not exists {tname} ({schema})"


def _dict_factory(cursor, row):
    d = {}
    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d


def _keyfn(cols):
    if len(cols) == 1:
        col = cols[0]
        return lambda r: r[col]
    return lambda r: [r[col] for col in cols]


def _delete(c, tname):
    c.cursor().execute(f'drop table if exists {tname}')


def _insert(c, tname, rs):
    rs = iter(rs)
    try:
        r0 = next(rs)
    except StopIteration:
        raise ValueError(f"No row to insert in {tname}") from None
    else:
        cols = list(r0)

        c.cursor().execute(_create_statement(tname, cols))
        istmt = _insert_statement(tname, r0)
        c.cursor().executemany(istmt, chain([r0], rs))


def _fetch(c, tname, cols):
    if cols:
        query = f"select * from {tname} order by {','.join(cols)}"
    else:
        query = f"select * from {tname}"

    yield from c.cursor().execute(query)


def _spy(it):
    val = next(it)
    return val, chain([val], it)


def _step(fn, krs1, krs2):
    empty = object()
    try:

        k1, rs1 = next(krs1)
        k2, rs2 = next(krs2)

        r1, rs1 = _spy(rs1)
        r2, rs2 = _spy(rs2)

        left = [c for c in list(r1) if c not in list(r2)]
        right = [c for c in list(r2) if c not in list(r1)]

        while True:
            if k1 == k2:
                yield from fn(rs1, rs2, left, right)
                k1 = k2 = empty
                k1, rs1 = next(krs1)
                k2, rs2 = next(krs2)
            elif k1 < k2:
                yield from fn(rs1, [], left, right)
                k1 = empty
                k1, rs1 = next(krs1)
            else:
                yield from fn([], rs2, left, right)
                k2 = empty
                k2, rs2 = next(krs2)

    except StopIteration:
        # unconsumed
        if k1 is not empty:
            yield from fn(rs1, [], left, right)
        if k2 is not empty:
            yield from fn([], rs2, left, right)

        for _, rs1 in krs1:
            yield from fn(rs1, [], left, right)
        for _, rs2 in krs2:
            yield from fn([], rs2, left, right)


# InstructionSet (It) to iter
def _rows2iter(obj):
    return obj.iter() if isinstance(obj, Rows) else iter(obj)


def _build_initgen(conn, tname, cols):
    def initgen():
        try:
            yield from _fetch(conn._dbconn, tname, cols)
        # in case conn._dbconn is either None or closed connection
        except (AttributeError, sqlite3.ProgrammingError):
            with _connect(conn._dbfile) as c:
                conn._dbconn = c
                yield from _fetch(c, tname, cols)
    return initgen


def _windowed(seq, n, step):
    if n < 0:
        raise ValueError('n must be >= 0')
    if n == 0:
        yield []
        return
    if step < 1:
        raise ValueError('step must be >= 1')

    window = deque(maxlen=n)
    i = n
    for _ in map(window.append, seq):
        i -= 1
        if not i:
            i = step
            yield list(window)

    size = len(window)
    if size == 0:
        return
    if size < n:
        yield list(window)
    elif 0 < i < min(step, n):
        yield list(window)[i:]


@contextmanager
def _connect(db):
    conn = sqlite3.connect(db)
    conn.row_factory = _dict_factory
    try:
        yield conn
    finally:
        # If users enter ctrl-c during the database commit,
        # db might be corrupted. (won't work anymore)
        with _delayed_keyboard_interrupts():
            conn.commit()
            conn.close()


@contextmanager
def _delayed_keyboard_interrupts():
    signal_received = []

    def handler(sig, frame):
        nonlocal signal_received
        signal_received = (sig, frame)
    # Do nothing but recording something has happened.
    old_handler = signal.signal(signal.SIGINT, handler)

    try:
        yield
    finally:
        # signal handler back to the original one.
        signal.signal(signal.SIGINT, old_handler)
        if signal_received:
            # do the delayed work
            old_handler(*signal_received)



from setuptools import setup

setup(
    name='tablemap',
    version='0.2.17',
    description='sql not knowing sql',
    long_description='',
    url='https://github.com/nalssee/tablemap.git',
    author='nalssee',
    author_email='jinisrolling@gmail.com',

    license='MIT',

)
